;; Generated package description from beacon.el  -*- no-byte-compile: t -*-
(define-package "beacon" "1.3.3.0.20190104.193148" "Highlight the cursor whenever the window scrolls" '((seq "2.14")) :keywords '("convenience") :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :url "https://github.com/Malabarba/beacon")
